# This module should calculate verification metrics at any time of the workflow
# It should implement different verification metrics:
# - FRPSS and RPSS
# - FCRPSS and CRPSS
# - enscorr
# - bias
# - reliability diagram
# - ask Carlos which decadal metrics he is currently using

source("modules/Skill/R/compute_quants.R")
source("modules/Skill/R/compute_probs.R")
source("modules/Skill/R/s2s.metrics.R")
source("modules/Saving/R/drop_dims.R")
## TODO: Remove with the next release of s2dv
source("modules/Skill/R/tmp/RPS.R")
source("modules/Skill/R/tmp/RPSS.R")
source("modules/Skill/R/tmp/GetProbs.R")
source("modules/Skill/R/tmp/RandomWalkTest.R")
source("modules/Skill/R/RPS_clim.R")
source("modules/Skill/R/CRPS_clim.R")

## TODO: Implement this in the future
## Which parameter are required?
# if (!("obs" %in% ls()) || is.null(obs)) {
#   error(logger,
#         "There is no object 'obs' in the global environment or it is NULL")
# }
# if (stream == "fcst" && (!("fcst" %in% ls()) || is.null(fcst))) {
#   error(logger,
#         "There is no object 'fcst' in the global environment or it is NULL")
# }
# if (!("hcst" %in% ls()) || is.null(hcst)) {
#   error(logger,
#         "There is no object 'hcst' in the global environment or it is NULL")
# }
# if (!("metric" %in% ls()) || is.null(metric)) {
#   warn(logger,
#        "Verification metric not found and it is set as 'EnsCorr'.")
#   metric <- 'EnsCorr'
# }
# if (metric %in% c('FRPSS', 'RPSS')) {
#   metric_fun <- "veriApply"
#   metric_method <- "FairRpss"
# } else if (metric %in% c("FCRPSS", "CRPSS")) {
#   metric_fun <- "veriApply"
# } else if (metric %in% c("EnsCorr", "EnsCor")) {
#   metric_fun <- "veriApply"
#   metric_method <- "EnsCorr"
# #...
# } else {
#   error(logger, "Unknown verification metric defined in the recipe.")
#   metric_fun <- 'NotFound'
# }
# info(logger, paste("#-------------------------- ", "\n",
#                    " running Skill module ", "\n",
#                    " it can call ", metric_fun ))

# compute_skill_metrics <- function(recipe, data$hcst, obs,
# 				  clim_data$hcst = NULL,
# 				  clim_obs = NULL) {
compute_skill_metrics <- function(recipe, data, agg = 'global') {
  
  # data$hcst: s2dv_cube containing the hindcast
  # obs: s2dv_cube containing the observations
  # recipe: auto-s2s recipe as provided by read_yaml
  
  ## TODO: Adapt time_dims to subseasonal case
  ## TODO: Add dat_dim
  ## TODO: Refine error messages
  ## TODO: Add check to see if anomalies are provided (info inside s2dv_cube)
  
  #   if (recipe$Analysis$Workflow$Anomalies$compute) {
  #     if (is.null(clim_data$hcst) || is.null(clim_obs)) {
  #       warn(recipe$Run$logger, "Anomalies have been requested in the recipe,
  #          but the climatologies have not been provided in the 
  #          compute_skill_metrics call. Be aware that some metrics like the 
  #          Mean Bias may not be correct.")
  #     }
  #   } else {
  #     warn(recipe$Run$logger, "Anomaly computation was not requested in the
  #        recipe. Be aware that some metrics, such as the CRPSS may not be 
  #        correct.")
  #   }
  time_dim <- 'syear'
  memb_dim <- 'ensemble'
  metrics <- tolower(recipe$Analysis$Workflow$Skill$metric)
  if (is.null(recipe$Analysis$ncores)) {
    ncores <- 1
  } else {
    ncores <- recipe$Analysis$ncores
  }
  if (is.null(recipe$Analysis$remove_NAs)) {
    na.rm = F
  } else {
    na.rm = recipe$Analysis$remove_NAs
  }
  if (is.null(recipe$Analysis$Workflow$Skill$cross_validation)) {
    warn(recipe$Run$logger,
         "cross_validation parameter not defined, setting it to FALSE.")
    cross.val <- FALSE
  } else {
    cross.val <- recipe$Analysis$Workflow$Skill$cross_validation
  }
  
  ################# Random Walk plotting function ##################
  
  randomwalkplot <- function(score1, score2, filename, plot_title){
    
    counts <- 0
    for (i in 1:length(score1)){
      if (score1[i] < score2[i]){
        counts[i] <- 1
      } else if (score1[i] > score2[i]){
        counts[i] <- -1
      } else {
        counts[i] <- 0
      }
    }
    counts <- cumsum(counts)
    
    ylim <- max(max(abs(counts)), 2*sqrt(0:length(score1)))

    png(filename)
    plot(x = 0:length(score1), y = 2*sqrt(0:length(score1)),
         xlim = c(0,length(score1)), ylim = c(-ylim, ylim),
         col = 'black', type = 'l', xlab = 'Years', ylab = 'Counts',
         main = plot_title )
    lines(x = 0:length(score1), y = -2*sqrt(0:length(score1)),
          col = 'black', type = 'l')
    lines(x = 1:length(score1), y = counts,
          col = 'red', type = 'l')
    text(x = 5, y = ylim, 'Model more skillful')
    text(x = 6, y = -ylim, 'Climatology more skillful')
    dev.off()
  }
  
  
  ##### Aggregated significance calculation ########
  
  lats.min <- recipe$Analysis$Region$latmin
  lats.max <- recipe$Analysis$Region$latmax
  lons.min <- recipe$Analysis$Region$lonmin
  lons.max <- recipe$Analysis$Region$lonmax
  ncores <- recipe$Analysis$ncores
  
  spatial.metrics <- strsplit(recipe$Analysis$Workflow$Skill$spatial_metrics$metric, ", | |,")[[1]]
  
  ## Calculate aggregated significance for RPSS
  if ('rps' %in% spatial.metrics) {

    sign_rpss_spatial_aggr <- NULL

    ## Loop over regions
    for (reg in names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions)){
      ## Subset region of experimental data

      sub.region <- recipe$Analysis$Workflow$Skill$spatial_metrics$regions[[reg]]

      aux_hcst <- SelBox(data = data$hcst$data,
                         lon = as.vector(data$hcst$coords$longitude),
                         lat = as.vector(data$hcst$coords$latitude),
                         region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                    sub.region[['lat.min']], sub.region[['lat.max']]),
                         londim = 'longitude', latdim = 'latitude', mask = NULL)

      ## Subset region of observational data
      aux_obs <- SelBox(data = data$obs$data,
                        lon = as.vector(data$obs$coords$longitude),
                        lat = as.vector(data$obs$coords$latitude),
                        region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                   sub.region[['lat.min']], sub.region[['lat.max']]),
                        londim = 'longitude', latdim = 'latitude', mask = NULL)

      ## Calculate RPS
      .get_probs <- s2dv:::.GetProbs
      rps_hcst <- Apply(data = list(exp = aux_hcst$data, obs = aux_obs$data),
                        target_dims = list(exp = c('syear','ensemble'), obs = 'syear'),
                        fun = s2dv:::.RPS, time_dim = 'syear', memb_dim = 'ensemble',
                        indices_for_clim = 1:as.numeric(dim(data$hcst$data)['syear']),
                        prob_thresholds = c(1/3, 2/3), cross.val = cross.val, ncores = ncores)$output1

      ## Calculate RPS climatology
      rps_clim <- Apply(data = aux_obs$data, target_dims = 'syear', output_dims = 'syear',
                        return_mean = F, fun = RPS_clim, prob_thresholds = c(1/3, 2/3),
                        cross.val = cross.val, ncores = ncores)$output1

      ## Spatially aggregated RPS
      rps_spatial_aggr <- WeightedMean(data = rps_hcst,
                                        lon = as.vector(aux_hcst$lon),
                                        lat = as.vector(aux_hcst$lat),
                                        region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                                   sub.region[['lat.min']], sub.region[['lat.max']]),
                                        londim = 'longitude',
                                        latdim = 'latitude',
                                        na.rm = F,
                                        ncores = ncores)

      ## Spatially aggregated RPS
      rps_clim_spatial_aggr <- WeightedMean(data = rps_clim,
                                            lon = as.vector(aux_hcst$lon),
                                            lat = as.vector(aux_hcst$lat),
                                            region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                                       sub.region[['lat.min']], sub.region[['lat.max']]),
                                            londim = 'longitude',
                                            latdim = 'latitude',
                                            na.rm = F,
                                            ncores = ncores)

      ##########################
      

      for (ft in recipe$Analysis$Time$ftime_min : recipe$Analysis$Time$ftime_max){
        
        filename <- paste0(recipe$Run$output_dir, '/RPS_randomwalk_sdate', substr(recipe$Analysis$Time$sdate, 1,2), '_ftime', ft, 
                           '_',names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions[reg]),'.png')
        
        plot_title <- paste0('RPS, ', recipe$Analysis$Variables$name, ', ',names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions[reg]),
                             ', sdate: ', month.name[as.numeric(substr(recipe$Analysis$Time$sdate, 1,2))],
                             ', ftime: ', ft)
        
        plot <- randomwalkplot(rps_spatial_aggr[,1,1,1,1,ft,1], rps_clim_spatial_aggr[,1,1,1,1,ft,1], filename, plot_title)

      }
      
      ##########################
      
      ## Random walk test between RPS aggr and RPS climatology aggr
      sign_rpss_spatial_aggr[[reg]] <- Apply(data = list(x = rps_spatial_aggr, y = rps_clim_spatial_aggr),
                                             target_dims = 'syear',
                                             fun = function(x, y){s2dv:::.RandomWalkTest(as.vector(x), as.vector(y))},
                                             ncores = ncores)

    }

    ## Save output : significance for each ftime per region
    saveRDS(sign_rpss_spatial_aggr, file = paste0(recipe$Run$output_dir,'/rpss_spatial_aggr_sign_',recipe$Analysis$Time$sdate,'.RDS'))
  }

  ## Calculate aggregated significance for CRPSS
  if ('crps' %in% spatial.metrics) {

    sign_crpss_spatial_aggr <- NULL

    ## Loop over regions
    for (reg in names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions)){
      ## Subset region of experimental data

      sub.region <- recipe$Analysis$Workflow$Skill$spatial_metrics$regions[[reg]]

      aux_hcst <- SelBox(data = data$hcst$data,
                         lon = as.vector(data$hcst$coords$longitude),
                         lat = as.vector(data$hcst$coords$latitude),
                         region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                    sub.region[['lat.min']], sub.region[['lat.max']]),
                         londim = 'longitude', latdim = 'latitude', mask = NULL)

      ## Subset region of observational data
      aux_obs <- SelBox(data = data$obs$data,
                        lon = as.vector(data$obs$coords$longitude),
                        lat = as.vector(data$obs$coords$latitude),
                        region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                   sub.region[['lat.min']], sub.region[['lat.max']]),
                        londim = 'longitude', latdim = 'latitude', mask = NULL)

      ## Calculate CRPS
      crps_hcst <- Apply(data = list(exp = aux_hcst$data, obs = aux_obs$data),
                         target_dims = list(exp = c('syear','ensemble'), obs = 'syear'),
                         fun = s2dv:::.CRPS, time_dim = 'syear', memb_dim = 'ensemble',
                         ncores = ncores)$output1
       
      ## Calculate CRPS climatology
      crps_clim <- Apply(data = aux_obs$data, target_dims = 'syear', return_mean = FALSE, # output_dims = 'syear',
                         clim.cross.val = cross.val, fun = CRPS_clim, ncores = ncores)$output1 

      ## Spatially aggregated CRPS
      crps_spatial_aggr <- WeightedMean(data = crps_hcst,
                                       lon = as.vector(aux_hcst$lon),
                                       lat = as.vector(aux_hcst$lat),
                                       region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                                  sub.region[['lat.min']], sub.region[['lat.max']]),
                                       londim = 'longitude',
                                       latdim = 'latitude',
                                       na.rm = F,
                                       ncores = ncores)

      ## Spatially aggregated CRPS
      crps_clim_spatial_aggr <- WeightedMean(data = crps_clim,
                                            lon = as.vector(aux_hcst$lon),
                                            lat = as.vector(aux_hcst$lat),
                                            region = c(sub.region[['lon.min']], sub.region[['lon.max']],
                                                       sub.region[['lat.min']], sub.region[['lat.max']]),
                                            londim = 'longitude',
                                            latdim = 'latitude',
                                            na.rm = F,
                                            ncores = ncores)
      
      for (ft in recipe$Analysis$Time$ftime_min : recipe$Analysis$Time$ftime_max){
        
        filename <- paste0(recipe$Run$output_dir, '/CRPS_randomwalk_sdate', substr(recipe$Analysis$Time$sdate, 1,2), '_ftime', ft, 
                           '_',names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions[reg]),'.png')
        
        plot_title <- paste0('CRPS, ', recipe$Analysis$Variables$name, ', ',names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions[reg]),
                             ', sdate: ', month.name[as.numeric(substr(recipe$Analysis$Time$sdate, 1,2))],
                             ', ftime: ', ft)
        
        plot <- randomwalkplot(crps_spatial_aggr[,1,1,1,1,ft,1], crps_clim_spatial_aggr[,1,1,1,1,ft,1], filename, plot_title)
       
      }
      
      ##########################

      ## Random walk test between CRPS aggr and CRPS climatology aggr
      sign_crpss_spatial_aggr[[reg]] <- Apply(data = list(x = crps_spatial_aggr, y = crps_clim_spatial_aggr),
                                             target_dims = 'syear',
                                             fun = function(x, y){s2dv:::.RandomWalkTest(as.vector(x), as.vector(y))},
                                             ncores = ncores)
    }

    ## Save output : significance for each ftime per region
    saveRDS(sign_crpss_spatial_aggr, file = paste0(recipe$Run$output_dir,'/crpss_spatial_aggr_sign_',recipe$Analysis$Time$sdate,'.RDS'))
  }

  
  ############ END Aggregated significance calculation #############
  
  skill_metrics <- list()
  for (metric in strsplit(metrics, ", | |,")[[1]]) {
    # Whether the fair version of the metric is to be computed
    if (metric %in% c('frps', 'frpss', 'bss10', 'bss90',
                      'fcrps', 'fcrpss')) {
      Fair <- T
    } else {
      Fair <- F
    }
    # Whether to compute correlation for the ensemble mean or for each member
    if (metric == 'corr') {
      memb <- T
    } else if (metric == 'enscorr') {
      memb <- F
    } 
    # Ranked Probability Score and Fair version
    if (metric %in% c('rps', 'frps')) {
      skill <- RPS(data$hcst$data, data$obs$data,
                   time_dim = time_dim,
                   memb_dim = memb_dim,
                   Fair = Fair,
                   cross.val = cross.val,
                   ncores = ncores)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
      rps_clim <- Apply(list(data$obs$data),
                        target_dims = c(time_dim, memb_dim),
                        RPS_clim)$output1
      rps_clim <- .drop_dims(rps_clim)
      skill_metrics[[paste0(metric, "_clim")]] <- rps_clim
      # Ranked Probability Skill Score and Fair version
    } else if (metric %in% c('rpss', 'frpss')) {
      skill <- RPSS(data$hcst$data, data$obs$data,
                    time_dim = time_dim,
                    memb_dim = memb_dim,
                    Fair = Fair,
                    cross.val = cross.val,
                    ncores = ncores)
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
      # Brier Skill Score - 10th percentile
    } else if (metric == 'bss10') {
      skill <- RPSS(data$hcst$data, data$obs$data,
                    time_dim = time_dim, 
                    memb_dim = memb_dim,
                    prob_thresholds = 0.1,
                    cross.val = cross.val,
                    Fair = Fair,
                    ncores = ncores)
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
      # Brier Skill Score - 90th percentile
    } else if (metric == 'bss90') {
      skill <- RPSS(data$hcst$data, data$obs$data,
                    time_dim = time_dim,
                    memb_dim = memb_dim,
                    prob_thresholds = 0.9,
                    cross.val = cross.val,
                    Fair = Fair,
                    ncores = ncores)
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
      # CRPS and FCRPS
    } else if (metric %in% c('crps', 'fcrps')) {
      skill <- CRPS(data$hcst$data, data$obs$data,
                    time_dim = time_dim,
                    memb_dim = memb_dim,
                    Fair = Fair,
                    ncores = ncores)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
      crps_clim <- Apply(list(data$obs$data), target_dims = time_dim,
                         fun = CRPS_clim, memb_dim = memb_dim)$output1
      crps_clim <- .drop_dims(crps_clim)
      skill_metrics[['crps_clim']] <- crps_clim
      # CRPSS and FCRPSS
    } else if (metric %in% c('crpss', 'fcrpss')) {
      skill <- CRPSS(data$hcst$data, data$obs$data,
                     time_dim = time_dim,
                     memb_dim = memb_dim,
                     Fair = Fair,
                     ncores = ncores)
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$crpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
    } else if (metric == 'rms') {
      source("https://earth.bsc.es/gitlab/es/s2dv/-/raw/master/R/RMS.R")
      hcst_mean <- Apply(list(data$hcst$data), target_dims = memb_dim,
                         fun = mean, na.rm = na.rm, ncores = ncores)$output1
      hcst_mean <- InsertDim(hcst_mean, pos = 1, len = 1, name = memb_dim)
      skill <- RMS(exp = hcst_mean, 
                   obs = data$obs$data, 
                   time_dim = time_dim, dat_dim = NULL, comp_dim = NULL,
                   limits = NULL, conf = FALSE, alpha = 0.05, ncores = ncores)
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rms
      # Mean bias (climatology) 
    } else if (metric == 'mean_bias') {
      ## TODO: Eliminate option to compute from anomalies
      # Compute from full field
      if ((!is.null(data$hcst.full_val)) && (!is.null(data$obs.full_val)) && 
          (recipe$Analysis$Workflow$Anomalies$compute)) {
        skill <- Bias(data$hcst.full_val$data, data$obs.full_val$data,
                      time_dim = time_dim,
                      memb_dim = memb_dim,
                      ncores = ncores)
      } else {
        skill <- Bias(data$hcst$data, data$obs$data,
                      time_dim = time_dim,
                      memb_dim = memb_dim,
                      ncores = ncores)
      }
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
      # Mean bias skill score
    } else if (metric == 'mean_bias_ss') {
      if ((!is.null(data$hcst.full_val)) && (!is.null(data$obs.full_val)) && 
          (recipe$Analysis$Workflow$Anomalies$compute)) {
        skill <- AbsBiasSS(data$hcst.full_val$data, data$obs.full_val$data,
                           time_dim = time_dim,
                           memb_dim = memb_dim,
                           ncores = ncores)
      } else {
        skill <- AbsBiasSS(data$hcst$data, data$obs$data,
                           time_dim = time_dim,
                           memb_dim = memb_dim,
                           ncores = ncores)
      }
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$biasSS
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
      # Ensemble mean correlation
    } else if (metric %in% c('enscorr', 'corr')) {
      ## TODO: Return significance
      ## TODO: Implement option for Kendall and Spearman methods?
      skill <- s2dv::Corr(data$hcst$data, data$obs$data, 
                          dat_dim = 'dat', 
                          time_dim = time_dim, 
                          method = 'pearson', 
                          memb_dim = memb_dim,
                          memb = memb,
                          conf = F,
                          pval = F,
                          sign = T,
                          alpha = 0.05,
                          ncores = ncores)
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$corr
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
    } else if (metric == 'rmsss') {
      # Compute RMSSS
      skill <- RMSSS(data$hcst$data, data$obs$data,
                     dat_dim = 'dat',
                     time_dim = time_dim,
                     memb_dim = memb_dim,
                     pval = FALSE,
                     sign = TRUE,
                     sig_method = 'Random Walk',
                     ncores = ncores)
      # Compute ensemble mean and modify dimensions
      skill <- lapply(skill, function(x) {
        .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rmsss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
    } else if (metric == 'enssprerr') {
      # Remove ensemble dim from obs to avoid veriApply warning
      obs_noensdim <- ClimProjDiags::Subset(data$obs$data, "ensemble", 1,
                                            drop = "selected")
      capture.output(
        skill <- easyVerification::veriApply(verifun = 'EnsSprErr',
                                             fcst = data$hcst$data,
                                             obs = obs_noensdim,
                                             tdim = which(names(dim(data$hcst$data))==time_dim),
                                             ensdim = which(names(dim(data$hcst$data))==memb_dim),
                                             na.rm = na.rm,
                                             ncpus = ncores)
      )
      remove(obs_noensdim)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
      # SpecsVerification metrics  
    } else if (grepl("specs", metric, fixed = TRUE)) {
      # Compute SpecsVerification version of the metrics
      ## Retain _specs in metric name for clarity
      metric_name <- (strsplit(metric, "_"))[[1]][1] # Get metric name
      if (!(metric_name %in% c('frpss', 'frps', 'bss10', 'bss90', 'enscorr',
                               'rpss'))) {
        warn(recipe$Run$logger,
             "Some of the requested SpecsVerification metrics are not available.")
      }
      capture.output(
        skill <- Compute_verif_metrics(data$hcst$data, data$obs$data,
                                       skill_metrics = metric_name,
                                       verif.dims=c("syear", "sday", "sweek"),
                                       na.rm = na.rm,
                                       ncores = ncores)
      )
      skill <- .drop_dims(skill)
      if (metric_name == "frps") {
        # Compute yearly mean for FRPS
        skill <- colMeans(skill, dims = 1)
      }
      skill_metrics[[ metric ]] <- skill
    }
  }
  info(recipe$Run$logger, "##### SKILL METRIC COMPUTATION COMPLETE #####")
  # Save outputs
  #NURIA: I think change the output_dir is a problem for future savings
  recipe$Run$output_dir <- paste0(recipe$Run$output_dir,
                                  "/outputs/Skill/")
  # Separate 'corr' from the rest of the metrics because of extra 'ensemble' dim
  if (recipe$Analysis$Workflow$Skill$save == 'all') {
    corr_metric_names <- grep("^corr", names(skill_metrics))
    if (length(corr_metric_names) == 0) {
      save_metrics(recipe = recipe, skill = skill_metrics,
                   data_cube = data$hcst, agg = agg)
    } else {
      # Save corr
      if (length(skill_metrics[corr_metric_names]) > 0) {
        save_corr(recipe = recipe, skill = skill_metrics[corr_metric_names],
                  data_cube = data$hcst)
      }
      # Save other skill metrics
      if (length(skill_metrics[-corr_metric_names]) > 0) {
        save_metrics(recipe = recipe, skill = skill_metrics[-corr_metric_names],
                     data_cube = data$hcst, agg = agg)
      }
    }
  }
  
  ######### Calculate temporarily aggregated significance  ######
  # 
  # ## Calculate aggregated significance for RPSS
  # if ('rps' %in% spatial.metrics) {
  #   
  #   sign_rpss_temporal_aggr <- NULL
  #   
  #   ## Loop over regions
  #   for (reg in names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions)){ 
  #     ## Subset region of experimental data
  #     
  #     sub.region <- recipe$Analysis$Workflow$Skill$spatial_metrics$regions[[reg]]
  #     
  #     rps_temporal_aggr <- SelBox(data = skill_metrics$rps, 
  #                                 lon = as.vector(data$hcst$coords$longitude),
  #                                 lat = as.vector(data$hcst$coords$latitude), 
  #                                 region = c(sub.region[['lon.min']], sub.region[['lon.max']],
  #                                            sub.region[['lat.min']], sub.region[['lat.max']]),
  #                                 londim = 'longitude', latdim = 'latitude', mask = NULL)
  #     
  #     ## Subset region of observational data
  #     rps_clim_temporal_aggr <- SelBox(data = skill_metrics$rps_clim, 
  #                                      lon = as.vector(data$obs$coords$longitude),
  #                                      lat = as.vector(data$obs$coords$latitude), 
  #                                      region = c(sub.region[['lon.min']], sub.region[['lon.max']],
  #                                                 sub.region[['lat.min']], sub.region[['lat.max']]),
  #                                      londim = 'longitude', latdim = 'latitude', mask = NULL)
  #     
  #     
  #     ## Random walk test between RPS aggr and RPS climatology aggr
  #     sign_rpss_temporal_aggr[[reg]] <- Apply(data = list(x = rps_temporal_aggr$data, y = rps_clim_temporal_aggr$data),
  #                                             target_dims = c('latitude', 'longitude'),
  #                                             fun = function(x, y){s2dv:::.RandomWalkTest(as.vector(x), as.vector(y))},
  #                                             ncores = ncores)$sign
  #     
  #   }  
  #   
  #   ## Save output : significance for each ftime per region
  #   saveRDS(sign_rpss_temporal_aggr, file = paste0(recipe$Run$output_dir,'/rpss_temporal_aggr_sign_',recipe$Analysis$Time$sdate,'.RDS')) 
  # }
  # 
  # 
  # ## Calculate aggregated significance for RPSS
  # if ('crps' %in% spatial.metrics) {
  #   
  #   sign_crpss_temporal_aggr <- NULL
  #   
  #   ## Loop over regions
  #   for (reg in names(recipe$Analysis$Workflow$Skill$spatial_metrics$regions)){ 
  #     ## Subset region of experimental data
  #     
  #     sub.region <- recipe$Analysis$Workflow$Skill$spatial_metrics$regions[[reg]]
  #     
  #     crps_temporal_aggr <- SelBox(data = skill_metrics$crps, 
  #                                 lon = as.vector(data$hcst$coords$longitude),
  #                                 lat = as.vector(data$hcst$coords$latitude), 
  #                                 region = c(sub.region[['lon.min']], sub.region[['lon.max']],
  #                                            sub.region[['lat.min']], sub.region[['lat.max']]),
  #                                 londim = 'longitude', latdim = 'latitude', mask = NULL)
  #     
  #     ## Subset region of observational data
  #     crps_clim_temporal_aggr <- SelBox(data = skill_metrics$crps_clim, 
  #                                      lon = as.vector(data$obs$coords$longitude),
  #                                      lat = as.vector(data$obs$coords$latitude), 
  #                                      region = c(sub.region[['lon.min']], sub.region[['lon.max']],
  #                                                 sub.region[['lat.min']], sub.region[['lat.max']]),
  #                                      londim = 'longitude', latdim = 'latitude', mask = NULL)
  #     
  #     
  #     ## Random walk test between RPS aggr and RPS climatology aggr
  #     sign_crpss_temporal_aggr[[reg]] <- Apply(data = list(x = crps_temporal_aggr$data, y = crps_clim_temporal_aggr$data),
  #                                             target_dims = c('latitude', 'longitude'),
  #                                             fun = function(x, y){s2dv:::.RandomWalkTest(as.vector(x), as.vector(y))},
  #                                             ncores = ncores)$sign
  #     
  #   }  
  #   
  #   ## Save output : significance for each ftime per region
  #   saveRDS(sign_crpss_temporal_aggr, file = paste0(recipe$Run$output_dir,'/crpss_temporal_aggr_sign_',recipe$Analysis$Time$sdate,'.RDS')) 
  # }
  
  
  # Return results
  return(skill_metrics)
  
}



############ END Calculate temporalily aggregated significance ###############



compute_probabilities <- function(recipe, data) {
  ## TODO: Do hcst and fcst at the same time 
  
  if (is.null(recipe$Analysis$ncores)) {
    ncores <- 1
  } else {
    ncores <- recipe$Analysis$ncores
  }
  
  if (is.null(recipe$Analysis$remove_NAs)) {
    na.rm = F
  } else {
    na.rm = recipe$Analysis$remove_NAs
  }
  
  named_probs <- list()
  named_probs_fcst <- list()
  named_quantiles <- list()
  
  if (is.null(recipe$Analysis$Workflow$Probabilities$percentiles)) {
    error(recipe$Run$logger, "Quantiles and probability bins have been 
          requested, but no thresholds are provided in the recipe.")
    stop()
  } else {
    for (element in recipe$Analysis$Workflow$Probabilities$percentiles) {
      # Parse thresholds in recipe
      thresholds <- sapply(element, function (x) eval(parse(text = x)))
      quants <- compute_quants(data$hcst$data, thresholds,
                               ncores = ncores,
                               na.rm = na.rm)
      probs <- compute_probs(data$hcst$data, quants,
                             ncores = ncores,
                             na.rm = na.rm)
      
      for (i in seq(1:dim(quants)['bin'][[1]])) {
        named_quantiles <- append(named_quantiles,
                                  list(ClimProjDiags::Subset(quants,
                                                             'bin', i)))
        names(named_quantiles)[length(named_quantiles)] <- paste0("percentile_",
                                                                  as.integer(thresholds[i]*100))
      }
      for (i in seq(1:dim(probs)['bin'][[1]])) {
        if (i == 1) {
          name_i <- paste0("prob_b", as.integer(thresholds[1]*100))
        } else if (i == dim(probs)['bin'][[1]]) {
          name_i <- paste0("prob_a", as.integer(thresholds[i-1]*100))
        } else {
          name_i <- paste0("prob_", as.integer(thresholds[i-1]*100), "_to_",
                           as.integer(thresholds[i]*100))
        }
        named_probs <- append(named_probs,
                              list(ClimProjDiags::Subset(probs,
                                                         'bin', i)))
        names(named_probs)[length(named_probs)] <- name_i
      }
      
      # Compute fcst probability bins
      if (!is.null(data$fcst)) {
        probs_fcst <- compute_probs(data$fcst$data, quants,
                                    ncores = ncores,
                                    na.rm = na.rm)
        
        for (i in seq(1:dim(probs_fcst)['bin'][[1]])) {
          if (i == 1) {
            name_i <- paste0("prob_b", as.integer(thresholds[1]*100))
          } else if (i == dim(probs_fcst)['bin'][[1]]) {
            name_i <- paste0("prob_a", as.integer(thresholds[i-1]*100))
          } else {
            name_i <- paste0("prob_", as.integer(thresholds[i-1]*100), "_to_",
                             as.integer(thresholds[i]*100))
          }
          named_probs_fcst <- append(named_probs_fcst,
                                     list(ClimProjDiags::Subset(probs_fcst,
                                                                'bin', i)))
          names(named_probs_fcst)[length(named_probs_fcst)] <- name_i
        }
      }
    }
    # Rearrange dimensions and return probabilities
    named_probs <- lapply(named_probs, function(x) {.drop_dims(x)})
    named_quantiles <- lapply(named_quantiles, function(x) {.drop_dims(x)})
    if (!is.null(data$fcst)) {
      fcst_years <- dim(data$fcst$data)[['syear']]
      named_probs_fcst <- lapply(named_probs_fcst, function(x) {.drop_dims(x)})
      #                           function(x) {Subset(x,
      #                                               along = 'syear', 
      #                                               indices = 1:fcst_years,
      #                                               drop = 'non-selected')})
      results <- list(probs = named_probs,
                      probs_fcst = named_probs_fcst,
                      percentiles = named_quantiles)
    } else {
      results <- list(probs = named_probs,
                      percentiles = named_quantiles)
    }
    
    info(recipe$Run$logger,
         "##### PERCENTILES AND PROBABILITY CATEGORIES COMPUTED #####")
    # Save outputs
    recipe$Run$output_dir <- paste0(recipe$Run$output_dir,
                                    "/outputs/Skill/")
    # Save percentiles
    if (recipe$Analysis$Workflow$Probabilities$save %in%
        c('all', 'percentiles_only')) {
      save_percentiles(recipe = recipe, percentiles = results$percentiles,
                       data_cube = data$hcst)
    }
    # Save probability bins
    if (recipe$Analysis$Workflow$Probabilities$save %in%
        c('all', 'bins_only')) {
      save_probabilities(recipe = recipe, probs = results$probs,
                         data_cube = data$hcst, type = "hcst")
      if (!is.null(results$probs_fcst)) {
        save_probabilities(recipe = recipe, probs = results$probs_fcst,
                           data_cube = data$fcst, type = "fcst")
      }
    }
    # Return results
    return(results)
  }
}

