

compute_statistics <- function(recipe, data){
  
  # data$hcst: s2dv_cube containing the hindcast
  # obs: s2dv_cube containing the observations
  # recipe: auto-s2s recipe as provided by read_yaml
  
  
  statistics <- tolower(recipe$Analysis$Workflow$Statistics$metrics)
  
  data_statistics <- list()
  
  for (stat in strsplit(statistics, ", | |,")[[1]]) {
    # Whether the fair version of the metric is to be computed
    if (stat %in% c('cov', 'covariance')) {
      
      ## Calculate mean
      mean_hcst <- Apply(data = data$hcst$data, target_dims = c('syear','ensemble'), fun = 'mean')$output1
      mean_obs <- Apply(data = data$obs$data, target_dims = c('syear','ensemble'), fun = 'mean')$output1
      
      temp_hcst <- NULL
      
      for(yr in 1:dim(data$hcst$data)['syear']){
        for(ens in 1:dim(data$hcst$data)['ensemble']){
          temp_hcst <-data$hcst$data[,,,,yr,,,,ens] - drop(mean_hcst)
          }
        }
    
      temp_obs <- NULL
      
      for(yr in 1:dim(data$obs$data)['syear']){
        for(ens in 1:dim(data$obs$data)['ensemble']){
          temp_obs <- data$obs$data[,,,,yr,,,,ens] - drop(mean_obs)
        }
      }
        
     product <- temp_hcst * temp_obs
     
     covariance <- sum(product)/(length(data$hcst$data)-1) ## doubt: hcst and obs have different lengths
     ## 0.00231112  
     
     #### CHECK ####
     
     ## remove ensemble dimension
     obs_data <- adrop(data$obs$data, drop = 9)
     ## Duplicate obs along hcst ensemble dimension
     obs_data <- InsertDim(data = obs_data, pos = 9, lendim = 25, name = 'ensemble')
     
     
     cov_test <- cov(x = data$hcst$data, y = obs_data, use = "everything", method = "pearson")
     ## [1] 3.773174
     cov_test_1 <- Apply(data = list(x= obs_data, y=data$hcst$data), target_dims = c('syear','ensemble'),
                         fun = function(x,y){cov(as.vector(x),as.vector(y), use = "everything", method = "pearson")})$output1
     
     data_statistics[[ stat ]] <- covariance
       
    } ## close if on cov
       
    
    if (stat %in% c('std', 'standard_deviation')) {
      
       ## Calculate standard deviation
       std_hcst <- Apply(data = data$hcst$data, target_dims = c('syear','ensemble'), fun = 'sd')$output1
       std_obs <- Apply(data = data$obs$data, target_dims = c('syear','ensemble'), fun = 'sd')$output1

       data_statistics[[ stat ]] <- list('std_hcst' = std_hcst, 'std_obs' = std_obs)
       
     } ## close if on std
  
  }
  
}  